# Copyright (C) 2009-2010 Red Hat, Inc.
#
# This program is free software; you can redistribute
# it and/or modify it under the terms of version 2 of the
# GNU General Public License as published by the
# Free Software Foundation.

"""Database Helpers used in luci."""

from pylons.i18n import ugettext as _
from sqlalchemy.orm.exc import NoResultFound

from luci.model import DBSession
from luci.lib.cluster_status import ClusterStatus
from luci.model.objects import Node, Cluster
from luci.lib.luci_tasks import luci_task_blocks_cluster

from luci.lib.ricci_defines import RICCI_BATCH_FAILURES_MAX
from luci.lib.ricci_communicator import RicciCommunicator
import luci.lib.ricci_queries as rq

import transaction
import logging
log = logging.getLogger(__name__)

def get_cluster_db_obj(cluster_name):
    db_obj = None
    try:
        db_obj = DBSession.query(Cluster).filter_by(name=cluster_name).one()
    except Exception, e:
        log.exception('Database object for cluster %s not found' % cluster_name)
    return db_obj

def get_cluster_node(cluster_name, node_name):
    db_obj = get_cluster_db_obj(cluster_name)
    if db_obj == None:
        return None
    for i in db_obj.nodes:
        if i.node_name == node_name:
            return i
    return None

def get_node_by_name(node_name):
    db_obj = None
    try:
        db_obj = DBSession.query(Node).filter_by(node_name=node_name).one()
    except Exception, e:
        log.exception('Database node object with name %s not found' % node_name)
    return db_obj

def get_node_by_host(node_host):
    db_obj = None
    try:
        db_obj = DBSession.query(Node).filter_by(hostname=node_host).one()
    except NoResultFound:
        log.debug('Database node object with host %s not found' % node_host)
    except Exception, e:
        log.exception('Database node object with host %s not found' % node_host)
    return db_obj

def get_agent_for_cluster(cluster_name):
    db_obj = get_cluster_db_obj(cluster_name)

    if db_obj is None:
        log.debug("Database object for cluster %s is missing" % cluster_name)
        return None

    # Try to use nodes that aren't busy first
    try:
        node_list = DBSession.query(Node).filter(Node.cluster.contains(db_obj)).order_by(Node.tasks != None).all()
    except:
        node_list = None

    if not node_list:
        node_list = db_obj.nodes

    node_list = db_obj.nodes
    if len(node_list) < 1:
        log.debug("No cluster nodes for cluster %s" % cluster_name)
        return None

    for node_obj in node_list:
        host = node_obj.hostname
        port = node_obj.port
        try:
            rc = RicciCommunicator(host, port)
            if rc is not None:
                if not rc.authed():
                    raise Exception, 'Not authenticated'
                return rc
        except Exception, e:
            log.exception('Error communicating with ricci agent at %s:%d: %s'
                % (host, port, e))

    log.error("Unable to communicate with ricci on any nodes in cluster %s"
        % cluster_name)
    return None

def get_model_for_cluster(cluster_name, rc=None):
    if rc is None:
        rc = get_agent_for_cluster(cluster_name)
        if rc is None:
            log.error("Unable to find a ricci agent for cluster %s"
                % cluster_name)
            return None

    try:
        from luci.lib.ClusterConf.ModelBuilder import ModelBuilder
        conf = rq.getClusterConf(rc)
        if conf is not None:
            model = ModelBuilder(None, conf, rc.cluster_version())
            return model
    except Exception, e:
        log.exception("Error getting cluster configuration for %s: %s"
            % (cluster_name, e))

    # Couldn't get the conf from any nodes
    return None

def get_cluster_status(rc):
    try:
        doc = rq.getClusterStatusBatch(rc)
    except Exception, e:
        log.exception('Error getting cluster status')
        return None
    return doc

def get_status_for_cluster(name, rc=None):
    if rc is None:
        rc = get_agent_for_cluster(name)

    if rc:
        status_xml = get_cluster_status(rc)
        if status_xml is None:
            log.debug("Error getting cluster status for %s" % name)
            return None
    else:
        status_xml = ''

    try:
        return ClusterStatus(status_xml)
    except Exception, e:
        log.exception("Error parsing status XML for cluster %s" % name)
        return None

def get_cluster_list():
    db_obj = None
    try:
        db_obj = DBSession.query(Cluster).all()
    except NoResultFound, e:
        log.error("No clusters found")
        return None
    except Exception, e:
        log.exception("Error accessing cluster database objects")
        return None

    return db_obj

def get_cluster_list_status(delegate=None):
    cluster_list = {}

    db_objs = get_cluster_list()
    for i in db_objs:
        cluster_name = i.name
        # Delegate, if it exists, will be an instance of
        #   luci.controllers.cluster.IndividualClusterController
        if delegate and delegate.name == cluster_name:
            status = delegate.get_status()
        else:
            rc = get_agent_for_cluster(cluster_name)
            if rc is None:
                status = ClusterStatus(None)
            else:
                status = get_status_for_cluster(cluster_name, rc)

        cluster_list[cluster_name] = {
            'status': status
            }
    return cluster_list

def get_cluster_list_full():
    cluster_list = {}

    db_objs = get_cluster_list()
    for i in db_objs:
        cluster_name = i.name
        rc = get_agent_for_cluster(cluster_name)
        if rc is None:
            cluster_list[cluster_name] = {
                'model': None,
                'status': ClusterStatus(None)
            }
            continue
        cluster_list[cluster_name] = {
            'model': get_model_for_cluster(cluster_name, rc),
            'status': get_status_for_cluster(cluster_name, rc)
        }
    return cluster_list

def create_cluster_obj(cluster_name, node_list):
    try:
        db_obj = Cluster(name=cluster_name, display_name=cluster_name, nodes=[])
        DBSession.add(db_obj)
        for node_host, node_port in node_list:
            node_obj = Node(node_name=node_host,
                            hostname=node_host,
                            ipaddr=node_host,
                            display_name=node_host,
                            port=node_port,
                            cluster=[db_obj])
            DBSession.add(node_obj)
        DBSession.flush()
        transaction.commit()
    except:
        log.exception("Error adding database objects for cluster: %s")
        DBSession.rollback()
        return False
        
    return True

def db_remove_cluster(cluster_name):
    db_obj = get_cluster_db_obj(cluster_name)
    if not db_obj:
        return False
    try:
        for i in db_obj.nodes:
            DBSession.delete(i)
        DBSession.delete(db_obj)
        DBSession.flush()
        transaction.commit()
    except:
        log.exception("Error removing cluster %s")
        DBSession.rollback()
        return False
    return True

def db_remove_cluster_node(cluster_name, node_name):
    node_to_remove = None
    try:
        cluster_obj = get_cluster_db_obj(cluster_name)
        for i in cluster_obj.nodes:
            if i.node_name == node_name or i.hostname == node_name:
                node_to_remove = i
                break

        if node_to_remove is not None:
            DBSession.delete(node_to_remove)
            DBSession.flush()
            transaction.commit()
            return True
    except:
        log.exception('Error removing node "%s" from the database' % node_name)
    return False

def get_cluster_task_status(cluster_name):
    task_status = {}

    db_obj = get_cluster_db_obj(cluster_name)
    if db_obj is None:
        return (False, task_status)

    tasks = db_obj.tasks
    if len(tasks) < 1:
        return (False, task_status)

    for t in tasks:
        task_bid = str(t.batch_id)
        task_type = t.task_type
        
        try:
            node_obj = t.node[0]
        except:
            log.exception('No node object for task %s' % task_bid)
            DBSession.delete(t)
            continue

        node_host = node_obj.hostname
        node_port = node_obj.port

        # return code = True denotes the cluster is blocked due to
        # a ricci job that is executing
        return_code = False
        try:
            rc = RicciCommunicator(node_host, node_port)
            ret = rc.batch_status(task_bid)
            if ret[0] is True:
                # job completed successfully
                task_status[task_bid] = {'status': ret[1]}
                DBSession.delete(t)
            elif ret[0] is False:
                # job is in progress
                task_status[task_bid] = {'status': t.status_msg}
                if luci_task_blocks_cluster(task_type):
                    return_code = True
                t.batch_failures = 0
            elif ret[0] == -1:
                # job completed with errors
                task_status[task_bid] = {'error': ret[1]}
                DBSession.delete(t)
            elif ret[0] == -2:
                # unable to retrieve the status of the batch job.
                # try again later.
                batch_failures = t.batch_failures + 1
                if batch_failures >= RICCI_BATCH_FAILURES_MAX:
                    # give up
                    task_status[task_bid] = {'error': _('Unable to retrieve the status for batch job %s after %d tries') % (task_bid, RICCI_BATCH_FAILURES_MAX)}
                    DBSession.delete(t)
                t.batch_failures = batch_failures
            else:
                # This should not happen. If it does, we should not allow it
                # to block the cluster.
                log.error("Unexpected ricci batch %s response code from %s: %s"
                    % (task_bid, node_host, ret[0]))
                task_status[task_bid] = {'error': _('Got unexpected return code %d when attempting to retrieve the status for batch job %s') % (ret[0], task_bid)}
                DBSession.delete(t)
        except:
            return_code = True
            task_status[task_bid] = {'status': _('Unable to retrieve the status for batch job %s on node %s. The node may temporarily unreachable.') % (task_bid, node_host)}
            continue

    DBSession.flush()
    transaction.commit()
    return (return_code, task_status)
