# Copyright (C) 2009-2010 Red Hat, Inc.
#
# This program is free software; you can redistribute
# it and/or modify it under the terms of version 2 of the
# GNU General Public License as published by the
# Free Software Foundation.

# -*- coding: utf-8 -*-
"""Utilities used in connection with HTML forms."""

from string import maketrans
# Imported, but not in module's header: base64

__all__ = ['FormUtils']


class FormUtils:
    """Methods useful for HTML forms, especially `values conversion mechanism'.

    `Values conversion mechanism' for HTML forms:

        (X)HTML forms uses 'name' attribute of some nested elements to
        use its value within triggered request. This value has to conform
        the lexical requirements that do not allow some characters within it.

        This situation is handled by a pair of function providing `values
        conversion mechanism':
          1) the original value is encoded in a safe way before it is used
             as the value of mentioned attribute
          2) in the request triggered by corresponding HTML form, these
             values are converted back to the original form

        The implementation uses base64 encoding with some other additions;
        references:
            [1] http://www.w3.org/TR/xhtml1/#C_8
            [2] http://www.w3.org/TR/html4/types.html#h-6.2
            [3] http://tools.ietf.org/html/rfc3548.html

    """

    #---------------------------------------------------------------------------
    # Values conversion mechanism.

    # First character has to be [A-Za-z].
    # Note: it's better when `straight' keys (i.e. not converted by following
    #       methods) of the request triggered by the HTML form do not start
    #       with this letter if this conversion mechanism is also used in that
    #       HTML form.
    STARTING_CHAR = u'L'
    
    B64_ORIG_CHARS = u'=+/'
    B64_CONV_CHARS = u'-_:'

    TRANS_TABLE_TO_CONV = maketrans(B64_ORIG_CHARS, B64_CONV_CHARS)
    TRANS_TABLE_FROM_CONV = maketrans(B64_CONV_CHARS, B64_ORIG_CHARS)


    @classmethod
    def string2Id(cls, orig_string):
        """Convert original string to the form usable within HTML form.
    
        See doc-string for the class.
    
        Keyword arguments:
            orig_string    Original string to convert.

        Return value(s):
            Encoded string.
    
        """
        from base64 import b64encode

        b64_string = b64encode(orig_string.encode('utf-8'))
        return cls.STARTING_CHAR + b64_string.translate(cls.TRANS_TABLE_TO_CONV)
    

    @classmethod
    def id2String(cls, id_string):
        """Receive the original string previously encoded by string2Id().

        See doc-string for the class.
    
        Keyword arguments:
            id_string    String to convert back to human readable format.

        Return value(s):
            String converted back from its encoded form.

        """
        from base64 import b64decode

        b64_string = str(id_string)[1:].translate(cls.TRANS_TABLE_FROM_CONV)
        return b64decode(b64_string).decode('utf-8')
    

    #---------------------------------------------------------------------------
    # Other useful methods.

    @staticmethod
    def compactString(long_string, max_length, append=u'...'):
        """For string longer than max_length, shorten it and append e.g. '...'.
    
        Keyword arguments:
            long_string    String to be `compacted'.
            max_length     Maximum tolerated length of the input string.
            append   What to append to the `compacted' string.
        
        Return value:
            `Compacted' string if the input one was longer than max_length,
            original string otherwise.
    
        """
        if len(long_string) > max_length:
            return long_string[:max_length] + append
        else:
            return long_string

