# Copyright (C) 2009-2010 Red Hat, Inc.
#
# This program is free software; you can redistribute
# it and/or modify it under the terms of version 2 of the
# GNU General Public License as published by the
# Free Software Foundation.

from pylons.i18n import ugettext as _

from luci.lib.ClusterConf.FailoverDomain import FailoverDomain
from luci.lib.ClusterConf.FailoverDomainNode import FailoverDomainNode
from luci.lib.ClusterConf.QuorumD import QuorumD
from luci.lib.ClusterConf.Heuristic import Heuristic
from luci.lib.ClusterConf.Method import Method

from luci.lib.db_helpers import create_cluster_obj
from luci.lib.ricci_communicator import RicciCommunicator
from luci.lib.ricci_helpers import send_batch_parallel
from luci.lib.ClusterConf.Totem import Totem

from validate_fence import validateFenceDevice, validateNewFenceDevice, validate_fenceinstance

from xml.dom import minidom

import logging
log = logging.getLogger(__name__)

def validate_fdom_prop_form(model, **kw):
    errors = []

    fdom_name = kw.get('name')
    if not fdom_name:
        return (False, {'errors': _('No failover domain name was given')})
    fdom = model.getFailoverDomainByName(fdom_name)
    if not fdom:
        return (False, {'errors': _('No failover domain named "%s" exists') % fdom_name})

    prioritized = kw.get('prioritized')
    if prioritized:
        fdom.addAttribute('ordered', '1')
    else:
        fdom.addAttribute('ordered', '0')
        
    restricted = kw.get('restricted')
    if restricted:
        fdom.addAttribute('restricted', '1')
    else:
        fdom.addAttribute('restricted', '0')

    nofailback = kw.get('nofailback')
    if nofailback:
        fdom.addAttribute('nofailback', '1')
    else:
        fdom.addAttribute('nofailback', '0')

    return (len(errors) < 1, {'errors': errors})

def validate_fdom_create_form(model, **kw):
    errors = []

    fdom_name = kw.get('fdom_name')
    if not fdom_name:
        errors.append(_('No name was given for this failover domain'))
        return (False, {'errors': errors})

    if model.getFailoverDomainByName(fdom_name):
        errors.append(_('A failover domain named "%s" already exists') % fdom_name)
        return (False, {'errors': errors})

    fdom = FailoverDomain()
    fdom.addAttribute('name', fdom_name)

    prioritized = kw.get('prioritized')
    if prioritized:
        fdom.addAttribute('ordered', '1')
    else:
        fdom.addAttribute('ordered', '0')
        
    restricted = kw.get('restricted')
    if restricted:
        fdom.addAttribute('restricted', '1')
    else:
        fdom.addAttribute('restricted', '0')

    nofailback = kw.get('nofailback')
    if nofailback:
        fdom.addAttribute('nofailback', '1')
    else:
        fdom.addAttribute('nofailback', '0')

    for i in model.getNodeNames():
        if kw.get('%s.check' % i):
            fdn = FailoverDomainNode()
            if prioritized:
                priority = kw.get('%s.priority' % i)
                if priority:
                    try:
                        priority = int(priority)
                        if priority < 1:
                            raise ValueError
                        fdn.addAttribute('priority', str(priority))
                    except ValueError:
                        errors.append(_('Invalid priority for failover domain member %s: %s') % (i, priority))
            fdn.addAttribute('name', i)
            fdom.addChild(fdn)

    model.addFailoverDomain(fdom)
    return (len(errors) < 1, {'errors': errors})

def validate_fdom_prop_settings_form(model, **kw):
    errors = []

    fdom_name = kw.get('name')
    if not fdom_name:
        return (False, {'errors': _('No failover domain name was given')})
    fdom = model.getFailoverDomainByName(fdom_name)
    if not fdom:
        return (False, {'errors': _('No failover domain named "%s" exists') % fdom_name})

    restricted = fdom.is_restricted()
    prioritized = fdom.is_prioritized()

    if not restricted and not prioritized:
        return (True, {})

    for i in model.getNodeNames():
        fdn = fdom.get_member_node(i)
        if kw.get('%s.check' % i):
            if not fdn:
                fdn = FailoverDomainNode()
                fdn.addAttribute('name', i)
                fdom.addChild(fdn)
            if prioritized:
                priority = kw.get('%s.priority' % i)
                if priority:
                    try:
                        priority = int(priority)
                        if priority < 1:
                            raise ValueError
                        fdn.addAttribute('priority', str(priority))
                    except ValueError:
                        errors.append(_('Invalid priority for failover domain member %s: %s') % (i, priority))
        else:
            if fdn:
                fdom.removeChild(fdn)
            
    return (len(errors) < 1, {'errors': errors})

def validate_qdisk_prop_form(model, **kw):
    errors = []
    use_qdisk = None

    qdp = model.getQuorumdPtr()

    quorumd = kw.get('quorumd')
    if quorumd == 'true':
        use_qdisk = True
    elif quorumd == 'false':
        use_qdisk = False

    if use_qdisk is None:
        return (False, {'errors': [  _('No qdisk value was given') ] })

    if use_qdisk is False:
        if qdp is not None:
            model.delQuorumd()
            return (True, {'flash': [ _('The updated quorum disk settings will not take effect until the cluster is restarted') ] })
        # No changes
        return (False, {'errors': [ _('No changes were made to the quorum disk settings') ] })

    interval = kw.get('interval')
    if interval:
        try:
            interval = int(interval)
        except ValueError:
            errors.append(_('An invalid qdisk interval value was given: %s') % interval)

    votes = kw.get('votes')
    if votes:
        try:
            votes = int(votes)
        except ValueError:
            errors.append(_('An invalid qdisk votes value was given: %s') % votes)

    tko = kw.get('tko')
    if tko:
        try:
            tko = int(tko)
        except ValueError:
            errors.append(_('An invalid qdisk TKO value was given: %s') % tko)

    min_score = kw.get('min_score')
    if min_score:
        try:
            min_score = int(min_score)
        except ValueError:
            errors.append(_('An invalid qdisk minimum score value was given: %s') % min_score)

    device = kw.get('device')
    label = kw.get('label')

    if not device and not label:
        errors.append(_('Neither a qdisk device nor label was given'))
    if device and label:
        errors.append(_('You may give either a qdisk device or label, but not both'))

    qd = QuorumD()
    if interval:
        qd.addAttribute('interval', str(interval))
    if votes:
        qd.addAttribute('votes', str(votes))
    if tko:
        qd.addAttribute('tko', str(tko))
    if min_score:
        qd.addAttribute('min_score', str(min_score))

    if device:
        qd.addAttribute('device', str(device))
    else:
        qd.addAttribute('label', str(label))

    for i in xrange(10):
        hprog = kw.get('heuristic%d:hprog' % i)
        hinterval = kw.get('heuristic%d:hinterval' % i)
        hscore = kw.get('heuristic%d:hscore' % i)
        htko = kw.get('heuristic%d:htko' % i)

        if not hprog and not hinterval and not hscore and not htko:
            continue

        if not hprog:
            errors.append(_('No qdisk program given for heuristic %d')
                % (i + 1))

        if hinterval:
            try:
                hinterval = int(hinterval)
            except ValueError:
                errors.append(_('Invalid qdisk interval given for heuristic %d: %s')
                    % (i + 1, hinterval))

        if hscore:
            try:
                hscore = int(hscore)
            except ValueError:
                errors.append(_('Invalid qdisk score given for heuristic %d: %s')
                    % (i + 1, hscore))

        if model.getClusterVersion() >= 3:
            if htko:
                try:
                    htko = int(htko)
                except ValueError:
                    errors.append(_('Invalid qdisk TKO value given for heuristic %d: %s')
                        % (i + 1, htko))

        qheur = Heuristic()
        qheur.addAttribute('program',  str(hprog))
        if hinterval:
            qheur.addAttribute('interval', str(hinterval))
        if hscore:
            qheur.addAttribute('score', str(hscore))
        if model.getClusterVersion() >= 3:
            if htko:
                qheur.addAttribute('tko', str(htko))
        qd.addChild(qheur)

    if len(errors) > 0:
        return (False, {'errors': errors })

    ret = (len(errors) < 1, {'errors': errors})
    if qdp is not None:
        model.delQuorumd()
    else:
        ret[1]['start_qdisk'] = True
    model.setQuorumd(qd)
    return ret

def validate_add_existing(**kw):
    errors = []

    cluster_name = kw.get('clustername')
    num_nodes = kw.get('num_nodes')

    if not cluster_name:
        errors.append(_("No cluster name was given"))

    if not num_nodes:
        errors.append(_("The number of cluster nodes was not given"))
    else:
        try:
            num_nodes = int(num_nodes)
        except ValueError:
            errors.append(_("An invalid number of cluster nodes was given: %s")
                % num_nodes)

    if len(errors) != 0:
        return (False, {'errors': errors })

    node_triples = []
    node_list = []
    for i in xrange(num_nodes):
        node_host = kw.get('node%d_host' % i)
        if not node_host:
            errors.append(_("No hostname was given for node %d") % i)
            continue

        node_port = kw.get('node%d_port' % i)
        if not node_port:
            errors.append(_("No port was given for node %s") % node_host)
        else:
            try:
                node_port = int(node_port)
                if node_port != node_port & 0xffff:
                    raise ValueError, node_port
            except ValueError:
                errors.append(_("An invalid port was given for node %s: %s")
                    % (node_host, kw.get('node%d_port' % i)))
        node_pass = kw.get('node%d_passwd' % i)
        if not node_pass:
            errors.append(_("No password was given for node %s") % node_host)
        node_triples.append(((node_host, node_port), RicciCommunicator.auth, [ node_pass ]))
        node_list.append((node_host, node_port))

    ret = send_batch_parallel(node_triples, 10)

    if len(errors) != 0:
        return (False, {'errors': errors })
    
    if create_cluster_obj(cluster_name, node_list) is not True:
        errors.append(_('Unable to create the database objects for cluster %s') % cluster_name)
    else:
        return (True, {'flash': [ _('Cluster "%s" was successfully added to luci') % cluster_name ]})
    return (len(errors) < 1, {'errors': errors})

def validate_cluster_config_form(model, **kw):
    errors = []
    conf_page = kw.get('page')
    if not conf_page:
        return (False, {'errors': [ _('No config page was given') ] })

    if conf_page == 'Fence':
        from luci.lib.ClusterConf.FenceXVMd import FenceXVMd

        fd = model.getFenceDaemonPtr()

        post_fail_delay = kw.get('post_fail_delay')
        if post_fail_delay:
            try:
                dummy = int(post_fail_delay)
            except Exception, e:
               errors.append(_('Invalid post fail delay: %s') % post_fail_delay)
            else:
                fd.setPostFailDelay(post_fail_delay)
       
        post_join_delay = kw.get('post_join_delay')
        if post_join_delay:
            try:
                dummy = int(post_join_delay)
            except Exception, e:
               errors.append(_('Invalid post join delay: %s') % post_join_delay)
            else:
                fd.setPostJoinDelay(post_join_delay)
        if model.getClusterVersion() == 2: 
            fence_xvmd = kw.get('fence_xvmd')
            if not fence_xvmd:
                model.delFenceXVM()
            else:
                xvm_obj = FenceXVMd()
                model.addFenceXVM(xvm_obj)
        return (len(errors) < 1, {'errors': errors})
    elif conf_page == 'General':
        config_version = kw.get('config_version')
        if config_version:
            try:
                cur_version = int(model.getClusterConfigVersion())
                config_version = int(config_version)
                if config_version < cur_version:
                    return (False, {'errors': [ _('The cluster configuration version entered must be greater than the current version (%d)') % cur_version ]})
            except ValueError:
                return (False, {'errors': [ _('Invalid value for configuration version: %s') % config_version ]})
            model.setClusterConfigVersion(str(config_version))
    elif conf_page == 'Network':
        multicast = kw.get('multicast')
        mcast_addr = kw.get('mcast_address')

        if multicast == "multicast":
            model.del_cluster_broadcast()
            model.del_cluster_multicast()
        elif multicast == "multicast_manual":
            model.del_cluster_broadcast()
            model.set_cluster_multicast(mcast_addr, None)
        elif multicast == "broadcast":
            model.set_cluster_broadcast()
        else:
            return (False, {'errors': [ _('Invalid value for multicast configuration: %s') % multicast]})

        totem = model.getTotemPtr()
        if totem is None:
            totem = model.addTotemPtr()

        join_timeout = kw.get('join_timeout')
        if join_timeout:
            try:
                dummy = int(join_timeout)
                if dummy < 1:
                    raise ValueError, 'Must be greater than 0'
            except:
                errors.append(_('Invalid value for totem join timeout : %s') % join_timeout)
            else:
                if int(join_timeout) != Totem.JOIN_TIMEOUT:
                    totem.addAttribute('join', join_timeout)
                else:
                    totem.removeAttribute('join')

        token_timeout = kw.get('token_timeout')
        if token_timeout:
            try:
                dummy = int(token_timeout)
                if dummy < 1:
                    raise ValueError, 'Must be greater than 0'
            except:
                errors.append(_('Invalid value for totem token timeout : %s') % token_timeout)
            else:
                if int(token_timeout) != Totem.TOKEN_TIMEOUT:
                    totem.addAttribute('token', token_timeout)
                else:
                    totem.removeAttribute('token')

        token_retransmits = kw.get('token_retransmits')
        if token_retransmits:
            try:
                dummy = int(token_retransmits)
                if dummy < 1:
                    raise ValueError, 'Must be greater than 0'
            except:
                errors.append(_('Invalid value for totem token retransmits before loss: %s') % token_retransmits)
            else:
                if int(token_retransmits) != Totem.TOKEN_RETRANSMITS_BEFORE_LOSS:
                    totem.addAttribute('token_retransmits_before_loss_const', token_retransmits)
                else:
                    totem.removeAttribute('token_retransmits_before_loss_const')

        consensus_timeout = kw.get('consensus_timeout')
        if consensus_timeout:
            try:
                dummy = int(consensus_timeout)
                if dummy < 1:
                    raise ValueError, 'Must be greater than 0'
            except:
                errors.append(_('Invalid value for totem consensus timeout : %s') % consensus_timeout)
            else:
                if int(consensus_timeout) != Totem.CONSENSUS_TIMEOUT:
                    totem.addAttribute('consensus', consensus_timeout)
                else:
                    totem.removeAttribute('consensus')

        return (len(errors) < 1, {'errors': errors})
    else:
        return (False, {'errors':
                        [ _('Unknown value %s for page was given') % conf_page ]})
    return (len(errors) < 1, {'errors': errors})

def validate_node_fence_level(fence_level, model, node, xml_str):
    errors = []
    nodename = node.getName()
    try:
        if not xml_str:
            raise Exception, 'no fence XML'
        doc = minidom.parseString(xml_str)
    except:
        return (False, {'errors': [ _('Unable to parse fence XML data for fence level %d') % fence_level ]})

    levels = node.getFenceLevels()
    try:
        method_id = levels[min(1, fence_level - 1)].getAttribute('name')
        if not method_id:
            raise Exception, 'No method ID'
        fence_method = Method()
        fence_method.addAttribute('name', str(method_id))
        levels[min(1, fence_level - 1)] = fence_method
    except Exception, e:
        method_id = str(fence_level)
        fence_method = Method()
        fence_method.addAttribute('name', str(method_id))

    forms = doc.getElementsByTagName('form')
    if len(forms) < 1:
        delete_target = None
        for l in levels:
            # delete the fence level
            if l.getAttribute('name') == method_id:
                delete_target = l
                break
        if delete_target is not None:
            try:
                node.getFenceNode().removeChild(delete_target)
            except Exception, e:
                return (False, {'errors': ['An error occurred while deleting fence method %s' % method_id ]})
        else:
            return (True, {'messages': ['No changes were made'] })

    form_hash = {}
    for i in forms:
        form_id = i.getAttribute('id')
        if not form_id:
            continue
        ielems = i.getElementsByTagName('input')
        if not ielems or len(ielems) < 1:
            continue

        dummy_form = {}

        for i in ielems:
            try:
                input_type = str(i.getAttribute('type'))
            except Exception, e:
                continue

            if not input_type or input_type == 'button':
                continue

            try:
                dummy_form[str(i.getAttribute('name'))] = str(i.getAttribute('value'))
            except Exception, e:
                log.exception("ielem")

        if len(dummy_form) < 1:
            continue

        if dummy_form.has_key('fence_instance'):
            try:
                parent = dummy_form['parent_fencedev']
            except:
                return (False, {'errors': [ 'Unable to determine what device the current instance uses' ]})

            try:
                form_hash[parent][1].append(dummy_form)
                del dummy_form['fence_instance']
            except Exception, e:
                return (False, {'errors': [ 'Unable to determine what device the current instance uses' ]})
        else:
            form_hash[form_id] = (dummy_form, list())

    fh_keys = form_hash.keys()
    fh_keys.sort()
    for i in fh_keys:
        fencedev_name = None
        fencedev_unknown = False

        try:
            fence_form, instance_list = form_hash[i]
        except Exception, e:
            log.exception("fence list")

        try:
            fence_type = fence_form['fence_type']
            if not fence_type:
                raise Exception, 'fence type is blank'
        except Exception, e:
            fence_type = None
            log.exception("fence type")

        if fence_form.has_key('existing_device'):
            try:
                fencedev_name = fence_form['name']
                if not fencedev_name.strip():
                    raise Exception, 'no fence name'
            except Exception, e:
                errors.append('You must provide a unique name for all fence devices')
                continue

            if fence_type is None:
                # An unknown fence device agent. Pull the data out of
                # the model and persist it and all instances.
                # All we care about is its name.
                fencedev_unknown = True
            else:
                if not fence_form.has_key('sharable'):
                    # If it's a shared fence device that already exists, the
                    # user could not have edited it so it's safe to pull the
                    # existing entry from the model. All we need is the
                    # device name, and nothing
                    # else needs to be done here.
                    #
                    # For an existing non-shared device update the device
                    # in the model, since the user could have edited it.
                    retcode, retmsg = validateFenceDevice(model, **fence_form)
                    if retcode != True:
                        errors.extend(retmsg)
                        continue
                    else:
                        fencedev_name = retmsg

                    # Add back the tags under the method block
                    # for the fence instance
                    if fence_type == 'fence_manual':
                        instance_list.append({'name': fencedev_name, 'nodename': nodename })
                    else:
                        instance_list.append({'name': fencedev_name })
        else:
            # The user created a new fence device.
            retcode, retmsg = validateNewFenceDevice(model, **fence_form)
            if retcode != True:
                errors.extend(retmsg)
                continue
            else:
                fencedev_name = retmsg

            # If it's not shared, we need to create an instance form
            # so the appropriate XML goes into the <method> block inside
            # <node><fence>. All we need for that is the device name.
            if not fence_form.has_key('sharable'):
                if fence_type == 'fence_manual':
                    instance_list.append({'name': fencedev_name, 'nodename': nodename })
                else:
                    instance_list.append({'name': fencedev_name })

        if fencedev_unknown is True:
            # Save any instances for this fence device.
            pass

        for inst in instance_list:
            retcode, retobj = validate_fenceinstance(fencedev_name, fence_type, **inst)
            if retcode != True:
                errors.extend(retobj)
                continue
            fence_method.addChild(retobj)

        fence_node = node.getFenceNode()
        found_target = False
        for idx in xrange(len(levels)):
            if levels[idx].getAttribute('name') == method_id:
                found_target = True
                break

        if found_target is False:
            # There's a fence block, but no relevant method
            # block
            fence_node.addChild(fence_method)

    return (len(errors) < 1, {'errors': errors})

def validate_node_fence_form(model, **kw):
    errors = []
    nodename = kw.get('node')
    if not nodename:
        return (False, {'errors': [ _('No node name was given') ]})

    try:
        node = model.retrieveNodeByName(nodename)
    except:
        return (False, {'errors': [ _('Unable to find node "%s" in the cluster configuration') % nodename ]})

    ret = validate_node_fence_level(1, model, node, kw.get('level1_xml'))
    if ret[0] is False:
        errors.extend(ret[1].get('errors'))
    ret = validate_node_fence_level(2, model, node, kw.get('level2_xml'))
    if ret[0] is False:
        errors.extend(ret[1].get('errors'))

    return (len(errors) < 1, {'errors': errors})
